
package com.cometway.jerminal;

import java.io.*;
import java.util.*;
import java.net.*;

import kinetoscope.util.*;

/**
 *	Copyright (c) 2000-2001, Comet Way, Inc.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, 
 * with or without modification, are permitted provided 
 * that the following conditions are met:
 * 
 *   - Redistributions of source code must retain the 
 *     above copyright notice, this list of conditions 
 *     and the following disclaimer. 
 *   - Redistributions in binary form must reproduce the 
 *     above copyright notice, this list of conditions 
 *     and the following disclaimer in the documentation 
 *     and/or other materials provided with the distribution.
 *   - Neither the name of Comet Way, Inc. nor the names of 
 *     its contributors may be used to endorse or promote 
 *     products derived from this software without specific 
 *     priorwritten permission. 
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 *  ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIEDWARRANTIES OF MERCHANTABILITY AND FITNESS 
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED.IN NO EVENT SHALL THE 
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 *  USE, DATA, OR PROFITS; ORBUSINESS INTERRUPTION) HOWEVER CAUSED AND 
 *  ON ANY THEORY OF LIABILITY, WHETHER INCONTRACT, STRICT LIABILITY, 
 *  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
 *  OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
 *  SUCH DAMAGE. 
 *
 * Revision history:
 */


/**
 * The Terminal classes are responsible for connecting the I/O streams from local
 * or remote JerminalSessions to the JerminalWindow and JerminalKeyListener. The OutputStream
 * of a JerminalSession needs to be connected to the JerminalWindow and the InputStream needs 
 * to be connected to the JerminalKeyListener associated with the JerminalWindow.
 * This terminal implements a JerminalListenerInterface to recieve notifications of the
 * JerminalSession status so that it can be relayed to appropriate classes.
 */
public class DumbTerminal implements Runnable, JerminalListenerInterface
{
	JerminalWindow window;
	JerminalKeyListener keys;
	Props props;

	Thread runThread;
	
	String remoteHost;

	boolean stopRunning;

	public DumbTerminal()
	{
		this(new String[0]);
	}

	public DumbTerminal(String[] args)
	{
		props = new Props();
		
		for(int x=0;x<args.length;x++) {
			int index = args[x].indexOf("=");
			if(index!=-1) {
				props.setProperty(args[x].substring(0,index),args[x].substring(index+1));
			}
			else {
				props.setProperty(args[x],"true");
			}
		}

		if(props.hasProperty("remote_host")) {
			remoteHost = props.getString("remote_host");
		}

		runThread = new Thread(this);
		runThread.setName("DumbTerminal");
		runThread.start();
	}



	/************** JerminalListenerInterface methods *********************/

	/**
	 * When the JerminalSession is executing a binary, the JerminalKeyListener must stop interpreting input.
	 */
	public void notifyRunningBinary()
	{
		//		System.out.println("Setting interpret to false");
		keys.setInterpret(false);
	}

	/**
	 * When the JerminalSession is finished executing a binary, the JerminalKeyListener must start interpreting input again.
	 */
	public void notifyFinishedRunningBinary()
	{
		//		System.out.println("Setting interpret to true");
		keys.setInterpret(true);
	}

	/**
	 * If the JerminalSession has been given an invalid command, the JerminalWindow will visually beep.
	 */
	public void notifyInvalidCommand()
	{
		window.visualBeep();
	}

	/**
	 * Not Implemented
	 */
	public void notifyExecutingCommand(String command)
	{

	}

	/**
	 * Not Implemented
	 */
	public void notifyFinishedExecutingCommand(String command)
	{

	}

	/** 
	 * This notification tells us we should clean house and shut down.
	 */
	public void notifyJerminalExiting()
	{
		stopRunning = true;
		if(window.jerminal.getOutputStream()!=null) {
			try {
				window.jerminal.getOutputStream().write(0);
				window.jerminal.getOutputStream().flush();
			}
			catch(Exception e) {
				e.printStackTrace();
			}
		}
		window.dispose();
		window.dispose();
		window.dispose();
		window = null;
		keys = null;
		runThread = null;
	}


	/**
	 * This is the method that sets everything up and reads from the JerminalSession's output stream
	 * and writes to the JerminalWindow.
	 */
	public void run()
	{
		try {
			InputStream in = null;
			OutputStream out = null;
			InputStream jIn = null;
			OutputStream jOut = null;

			if(remoteHost == null) {
				in = new PipedInputStream();
				jOut = new PipedOutputStream((PipedInputStream)in);
				jIn = new PipedInputStream();
				out = new PipedOutputStream((PipedInputStream)jIn);
			}
			else {
				Socket s = new Socket(remoteHost,props.getInteger("remote_port"));
				in = s.getInputStream();
				out = s.getOutputStream();
			}

			window = new JerminalWindow(props);

			keys = new EmacsKeyListener(window);
			window.keyListener = keys;
			window.addKeyListener(window.keyListener);

			Props p = new Props();
			if(props.hasProperty("jerminal_props")) {
				p = p.loadProps(props.getString("jerminal_props"));
			}
			//			if(props.hasProperty("use_default_class_aliases")) {
			p.setProperty("use_default_aliases","true");
				//			}
			p.setProperty("JerminalWindow",window);
			//			p.setProperty("JerminalWindow_PROPS",props);
			if(remoteHost==null) {
				window.jerminal = new JerminalSession(jIn,jOut,p,true,true,true);
				window.jerminalIn = in;
				window.jerminalOut = out;
				window.jerminal.listeners.addElement(this);
				window.jerminal.start();
			}
			else {
				window.jerminal = new JerminalSession(null,null,p,true,true,true);
				window.jerminalIn = in;
				window.jerminalOut = out;
			}

			if(props.hasProperty("use_std_error")) {
				System.setErr(new PrintStream(jOut));
			}
			if(props.hasProperty("use_std_out")) {
				System.setOut(new PrintStream(jOut));
			}

			try {
				while(!stopRunning) {
					boolean refreshAll = false;
					byte[] buffer = new byte[window.getMaxCharWidth(window.fixedFontWidth)*5];
					int bytesRead = in.read(buffer);
					String line = "";
					for(int x=0;x<bytesRead;x++) {
						int i = (int)buffer[x];
						char c = (char)i;
						//					int i = in.read();
						//					char c = (char)i;
						if(i==10 || i==13) {
							window.insertAtCursor(line);
							window.moveCursorX(line.length());
							line = "";
							window.makeNewLine();
							refreshAll = true;
						}
						else if(i==9) {
							//						String old = window.text[window.cursorY];
							//						window.moveCursorX(2);
							window.insertAtCursor(line);
							window.moveCursorX(line.length());
							line = "";
							window.insertTab();
							//						refreshAll = !keys.lineDiff(old,window.text[window.cursorY]);
						}
						else {
							//						String old = window.text[window.cursorY];
							//							window.insertAtCursor(c);
							//							window.moveCursorX(1);
							//						refreshAll = !keys.lineDiff(old,window.text[window.cursorY]);
							line = line+c;
						}
					}
					window.insertAtCursor(line);
					window.moveCursorX(line.length());
					window.refresh(true);
				}
			}
			catch(Exception e) {
				e.printStackTrace();
			}

		}
		catch(Exception e) {
			e.printStackTrace();
		}
	}

	
	public static void main(String[] args)
	{
		if(args.length>0 && args[0].equals("-listfonts")) {
			java.awt.Toolkit kit = java.awt.Toolkit.getDefaultToolkit();
			String[] fonts = kit.getFontList();
			for(int x=0;x<fonts.length;x++) {
				System.out.println(fonts[x]);
			}
			System.exit(0);
		}
		else {			
			DumbTerminal dummy = new DumbTerminal(args);
		}
	}

}
