package com.cometway.jerminal;

import java.util.*;
import java.io.*;
import java.lang.reflect.*;

import kinetoscope.util.*;

import com.cometway.jerminal.commands.JerminalCommand;

/**
 *	Copyright (c) 2000-2001, Comet Way, Inc.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, 
 * with or without modification, are permitted provided 
 * that the following conditions are met:
 * 
 *   - Redistributions of source code must retain the 
 *     above copyright notice, this list of conditions 
 *     and the following disclaimer. 
 *   - Redistributions in binary form must reproduce the 
 *     above copyright notice, this list of conditions 
 *     and the following disclaimer in the documentation 
 *     and/or other materials provided with the distribution.
 *   - Neither the name of Comet Way, Inc. nor the names of 
 *     its contributors may be used to endorse or promote 
 *     products derived from this software without specific 
 *     priorwritten permission. 
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 *  ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIEDWARRANTIES OF MERCHANTABILITY AND FITNESS 
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED.IN NO EVENT SHALL THE 
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 *  USE, DATA, OR PROFITS; ORBUSINESS INTERRUPTION) HOWEVER CAUSED AND 
 *  ON ANY THEORY OF LIABILITY, WHETHER INCONTRACT, STRICT LIABILITY, 
 *  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
 *  OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
 *  SUCH DAMAGE. 
 *
 * Revision history:
 */


/**
 * The JerminalSession. This class contains the parse, the interpreter, and the necessary session
 * variables that tracks the state of a session. 
 *
 */
public class JerminalSession implements Runnable
{
	/** Setting this to true will cause the JerminalSession to stop accepting commands and its thread to stop. */
	public boolean stopRunning;

	int sessionCounter = 0;

	PrintStream localPrintStream;
	/** This is the Hashtable of cached commands. This Hashtable can be cleared by the rehash command. */
	public Hashtable cachedCommands;

	boolean fileAccess;
	boolean binaryAccess;
	boolean streamAccess;

	Thread runThread;

	InputStream in;
	OutputStream out;
	Props props;

	Vector commandPrefix;

	boolean showStackTracesOnJerm = true;

	/** This is the Vector which contains this sessions command history */
	public Vector commandHistory;

	/** This is the Hashtable which tracks the classname aliases set by the aliasClass command */
	public Hashtable classAliases;

	/** This is the Vector of JerminalListenerInterface listening to this session. */
	public Vector listeners;



	/**
	 * This instantiates and initializes a JerminalSession. 
	 * @param in The InputStream which the JerminalSession is to read its commands from
	 * @param out The OutputStream which the JerminalSession is to write the results to.
	 * @param p The Props Object which will hold all the environment variables.
	 * @param allowFileAccess If set to FALSE, file access is denied in this session.
	 * @param allowBinaryAccess If set to FALSE, access to binaries on the local system is denied in this session.
	 */
	public JerminalSession(InputStream in, OutputStream out, Props p, boolean allowFileAccess, boolean allowBinaryAccess)
	{
		this.in = in;
		this.out = out;

		localPrintStream = new PrintStream(out);

		props = p;
		fileAccess = allowFileAccess;
		binaryAccess = allowBinaryAccess;

		stopRunning = false;
		listeners = new Vector();
		initProps();
	}

	/**
	 * This instantiates and initializes a JerminalSession. 
	 * @param in The InputStream which the JerminalSession is to read its commands from
	 * @param out The OutputStream which the JerminalSession is to write the results to.
	 * @param p The Props Object which will hold all the environment variables.
	 * @param allowFileAccess If set to FALSE, file access is denied in this session.
	 * @param allowBinaryAccess If set to FALSE, access to binaries on the local system is denied in this session.
	 * @param allowStreamAccess If set to TRUE, the io streams will be handed off to the processes that are spawned.
	 */
	public JerminalSession(InputStream in, OutputStream out, Props p, boolean allowFileAccess, boolean allowBinaryAccess, boolean allowStreamAccess)
	{
		this.in = in;
		this.out = out;

		localPrintStream = new PrintStream(out);

		props = p;
		fileAccess = allowFileAccess;
		binaryAccess = allowBinaryAccess;
		streamAccess = allowStreamAccess;

		stopRunning = false;
		listeners = new Vector();
		initProps();
	}


	/**
	 * This adds a JerminalListener to this JerminalSession
	 */
	public void addListener(JerminalListenerInterface listener)
	{
		if(!listeners.contains(listener)) {
			listeners.addElement(listener);
		}
	}

	/**
	 * This removes a JerminalListener.
	 */
	public void removeListener(JerminalListenerInterface listener)
	{
		listeners.removeElement(listener);
	}

	/**
	 * This method returns the Props object which holds all the environment variables.
	 * @return The Props object which contains all the environment variables of the current session.
	 */
	public Props getProps()
	{
		return(props);
	}

	/**
	 * This starts the JerminalSession on its own Thread of execution.
	 */
	public void start()
	{
		runThread = new Thread(this);
		runThread.setPriority(Thread.MIN_PRIORITY);
		runThread.setName("Jerminal");

		runThread.start();
	}

	/**
	 * This initializes the default evironment variables.
	 */
	public void initProps()
	{
		commandPrefix = new Vector();
		commandPrefix.addElement("com.cometway.jerminal.commands.");

		props.setDefault("home_directory","/usr/tmp");
		props.setDefault("prompt","Jerminal(@c)> ");
		props.setDefault("ignore_string","#");
		props.setDefault("pwd",new File(props.getString("home_directory")));
		props.setDefault("path_separator_char","/");
		props.setDefault("TERM","vt100");

		if(props.hasProperty("command_prefix_list")) {
			commandPrefix = commaToVector(props.getString("command_prefix_list"));
		}

		commandHistory = new Vector();

		if(classAliases==null) {
			classAliases = new Hashtable();
		}
		if(props.hasProperty("use_default_aliases")) {
			classAliases.put("String","java.lang.String");
			classAliases.put("Object","java.lang.Object");
			classAliases.put("StringBuffer","java.lang.StringBuffer");

			classAliases.put("Date","java.util.Date");
			classAliases.put("Vector","java.util.Vector");
			classAliases.put("Hashtable","java.util.Hashtable");

			classAliases.put("Props","kinetoscope.util.Props");
			classAliases.put("Pair","kinetoscope.util.Pair");
			classAliases.put("StringTools","kinetoscope.util.StringTools");
			classAliases.put("DebugTools","kinetoscope.util.DebugTools");
			classAliases.put("HTTPLoader","kinetoscope.net.HTTPLoader");

			classAliases.put("JAK","com.cometway.jak.JAK");
			classAliases.put("JerminalSession","com.cometway.jerminal.JerminalSession");
		}
	}
			
	/**
	 * This method is the prompt/read/evaluate cycle of the JerminalSession interpreter.
	 */ 
	public void run()
	{
		try {
			localPrintStream.println("Welcome to Jerminal v0.2\n");
		}
		catch(Exception e) {
			error("Could not write to OutputStream: ",e);
		}

		while(!stopRunning) {
			try {
				prompt();

				String line = "";
				int inChar = in.read();
				char c = (char)inChar;
				while(c!='\n' && inChar!=-1) {
					if(props.hasProperty("local_echo")) {
						localPrintStream.print(c);
					}
					line = line+c;
					inChar = in.read();
					c = (char)inChar;
					//					System.out.println("inChar = "+inChar+", c = "+c);
				}

				// This may be a little dodgy
				while(inChar==-1) {
					in = null;
					out = null;
					localPrintStream = null;
						
					return;
				}

				if(props.hasProperty("local_echo")) {
					localPrintStream.println();
				}

				// this part deals with the command history
				if(line.trim().startsWith("!")) {
					line = getHistoryCommand(line.trim().substring(1));
					if(line==null) {
						localPrintStream.println("Command not found");
						continue;
					}
					localPrintStream.println("! -> "+line);
				}
				else {
					if(line.trim().length()>0) {
						commandHistory.insertElementAt(line,0);
						if(commandHistory.size()>props.getInteger("history_size")) {
							commandHistory.removeElementAt(props.getInteger("history_size"));
						}
					}
				}

				Object rval = evaluate(line);
				if(rval!=null) {
					localPrintStream.print(rval.toString());
					localPrintStream.flush();
				}
			}
			catch(Exception e) {
				error("Exception in run, ",e);
			}
		}

		runThread = null;

		JerminalListenerInterface listener = null;
		for(int x=0;x<listeners.size();x++) {
			try {
				listener = (JerminalListenerInterface)listeners.elementAt(x);
				listener.notifyJerminalExiting();
			}
			catch(Exception e) {;}
		}
	}

	/**
	 * This method takes a String pattern and matches it with all the commands in the command history.
	 * @param s This is the pattern which to match the commands. If blank or null, the most recent command is returned
	 * @return Returns the most recent command that starts with the String parameter.
	 */
	public String getHistoryCommand(String s)
	{
		String rval = null;

		if(s==null || s.length()==0) {
			rval = commandHistory.elementAt(0).toString();
		}
		else {
			for(int x=0;x<commandHistory.size();x++) {
				String line = (String)commandHistory.elementAt(x);
				if(line.startsWith(s)) {
					rval = line;
					break;
				}
			}
		}

		return(rval);
	}

	/**
	 * This method takes a String alias and unaliases it to the Classname.
	 * @param alias This is the alias of the classname set by the aliasClass command
	 * @return Returns the class name which is aliased by the [alias] parameter.
	 */
	public String getAliasedClass(String alias)
	{
		String rval = alias;
		if(classAliases.containsKey(alias)) {
			rval = (String)classAliases.get(alias);
		}
		return(rval);
	}

	/**
	 * @return Returns whether file access is allowed in this session of the Jerminal
	 */
	public boolean allowFileAccess()
	{
		return(fileAccess);
	}

	/**
	 * @return Returns whether binary access is allowed in this session of the Jerminal
	 */
	public boolean allowBinaryAccess()
	{
		return(binaryAccess);
	}

	/**
	 * @return Returns the OutputStream which this Jerminal session is writing its results to.
	 */
	public OutputStream getOutputStream()
	{
		return(out);
	}

	/**
	 * @return Returns the InputStream which this Jerminal session is reading its commands from.
	 */
	public InputStream getInputStream()
	{
		return(in);
	}



	/**
	 * This method prints a prompt and performs the necesary maintenance on the session
	 * and the prompt. The prompt can have the appropriate codes: '@c'= the session counter,
	 * '@t'= current time (am/pm), '@T'= current time military, '@D'= current working directory
	 * as the full path, '@d'= current working directory with home replaced with ~
	 */
	protected void prompt()
	{
		if(localPrintStream==null) {
			try {
				localPrintStream = new PrintStream(out);
			}
			catch(Exception e) {;}
		}
		if(cachedCommands == null) {
			cachedCommands = new Hashtable();
		}
		if(classAliases == null) {
			classAliases = new Hashtable();
		}
		if(!props.hasProperty("home_directory")) {
			props.setProperty("home_directory","/usr/tmp");
		}
		if(!props.hasProperty("pwd")) {
			props.setProperty("pwd",new File(props.getString("home_directory")));
		}
		if(!props.hasProperty("jerminal")) {
			props.setProperty("jerminal",this);
		}
		if(!props.hasProperty("history_size")) {
			props.setProperty("history_size","50");
		}

		String prompt = props.getString("prompt");

		if(prompt.length()==0) {
			prompt = "Jerminal (@c)> ";
		}

		// check for session counter replacements
		while(prompt.indexOf("@c")!=-1) {
			int index = prompt.indexOf("@c");
			prompt = prompt.substring(0,index) + sessionCounter + prompt.substring(index+2);
		}
		
		// check for time replacements
		while(prompt.indexOf("@T")!=-1) {
			Calendar cal = Calendar.getInstance();
			int index = prompt.indexOf("@T");
			String timeString = cal.get(cal.HOUR_OF_DAY)+":"+cal.get(cal.MINUTE);
			prompt = prompt.substring(0,index) + timeString + prompt.substring(index+2);
		}
		while(prompt.indexOf("@t")!=-1) {
			Calendar cal = Calendar.getInstance();
			int index = prompt.indexOf("@t");
			String timeString = "";
			if(cal.get(cal.HOUR_OF_DAY)>12) {
				if(cal.get(cal.MINUTE)<10) {
					timeString = (cal.get(cal.HOUR_OF_DAY)-12)+":0"+cal.get(cal.MINUTE)+"pm";
				}
				else {
					timeString = (cal.get(cal.HOUR_OF_DAY)-12)+":"+cal.get(cal.MINUTE)+"pm";
				}
			}
			else if(cal.get(cal.HOUR_OF_DAY)==0) {
				if(cal.get(cal.MINUTE)<10) {
					timeString = "12:0"+cal.get(cal.MINUTE)+"am";
				}
				else {
					timeString = "12:"+cal.get(cal.MINUTE)+"am";
				}
			}
			else {
				if(cal.get(cal.MINUTE)<10) {
					timeString = (cal.get(cal.HOUR_OF_DAY))+":0"+cal.get(cal.MINUTE)+"am";
				}
				else {
					timeString = (cal.get(cal.HOUR_OF_DAY))+":"+cal.get(cal.MINUTE)+"am";
				}
			}
			prompt = prompt.substring(0,index) + timeString + prompt.substring(index+2);
		}

		// check for directory replacements
		while(prompt.indexOf("@D")!=-1) {
			int index = prompt.indexOf("@D");
			prompt = prompt.substring(0,index) + props.getString("pwd") + prompt.substring(index+2);
		}
		while(prompt.indexOf("@d")!=-1) {
			int index = prompt.indexOf("@d");
			String pwd = props.getString("pwd");
			String home = props.getString("home_directory");
			if(pwd.indexOf(home)!=-1) {
				if(home.charAt(home.length()-1)=='/' || home.charAt(home.length()-1)=='\\') {
					home = home.substring(0,home.length()-1);
				}
				pwd = pwd.substring(0,pwd.indexOf(home))+"~"+pwd.substring(pwd.indexOf(home)+home.length());
			}
			prompt = prompt.substring(0,index)+pwd+prompt.substring(index+2);
		}


		try {
			localPrintStream.print(prompt);
			localPrintStream.flush();
		}
		catch(Exception e) {
			error("Could not write prompt ",e);
		}

		sessionCounter++;
	}




	/**
	 * This method evaluates a command string. The syntax is [operator] [operands]...
	 * The operands can be evaluated using the '%' character. The result of the evaluation
	 * is returned as an Object. The operators are classnames in the 'command_prefix_list'
	 * variable. By default, the com.cometway.jerminal.commands package is automatically
	 * added to this list. If the given command is not found and binaries access is allowed,
	 * this method will attempt to execute the binary denoted by the command.
	 *
	 * @param command The Command String that needs to be evaluated.
	 * @return Returns the result of that evaluation, or the result of the 'evaluate' method call on the command class.
	 */
	public Object evaluate(String command)
	{
		// Questionable... do I need to do this?
		command = command.trim();

		Object rval = new String("Invalid command: "+command+"\n");
		
		if(props.hasProperty("ignore_string")) {
			if(command.startsWith(props.getString("ignore_string"))) {
				return(null);
			}
		}

		if(command.length() == 0) {
			return(null);
		}
			
		String operator = null;
		Vector operands = null;
		
		int index = 0;
		char c = command.charAt(index);
		while(!Character.isWhitespace(c)) {
			index ++;
			if(index >= command.length()) {
				index = command.length();
				break;
			}
			c = command.charAt(index);
		}

		operator = command.substring(0,index);
		operands = parseArgs(command.substring(index).trim());

		if((operator == null) || (operator.trim().length()==0)) {
			return(rval);
		}
		
		try {
			JerminalCommand jcommand = null;
			if(cachedCommands.containsKey(operator)) {
				jcommand = (JerminalCommand)cachedCommands.get(operator);
			}
			else {
				Class commandClass = null;
				for(int x=0;x<commandPrefix.size();x++) {
					try {
						String prefix = (String)commandPrefix.elementAt(x);
						commandClass = Class.forName(prefix+operator);
						break;
					}
					catch(Exception es) {
						;
					}
				}
				if(commandClass == null) {
					//					error("Command not found: "+operator,null);
				}
				else {
					Class[] params = { Class.forName("com.cometway.jerminal.JerminalSession") };
					Constructor con = commandClass.getConstructor(params);
					Object[] o = { this };

					jcommand = (JerminalCommand)con.newInstance(o);
				}
				
				if(jcommand!=null) {
					cachedCommands.put(operator,jcommand);
				}
			}

			if(jcommand!=null) {
				JerminalListenerInterface listener = null;
				for(int x=0;x<listeners.size();x++) {
					try {
						listener = (JerminalListenerInterface)listeners.elementAt(x);
						listener.notifyExecutingCommand(jcommand.toString());
					}
					catch(Exception e) {;}
				}
				rval = jcommand.evaluate(operands);
				for(int x=0;x<listeners.size();x++) {
					try {
						listener = (JerminalListenerInterface)listeners.elementAt(x);
						listener.notifyFinishedExecutingCommand(jcommand.toString());
					}
					catch(Exception e) {;}
				}
			}
			else {
				if(binaryAccess) {
					String commandString = operator;
					for(int x=0;x<operands.size();x++) {
						commandString = commandString+" "+operands.elementAt(x);
					}
					JerminalListenerInterface listener = null;
					for(int x=0;x<listeners.size();x++) {
						try {
							listener = (JerminalListenerInterface)listeners.elementAt(x);
							listener.notifyRunningBinary();
						}
						catch(Exception e) {;}
					}
					if(!streamAccess) {
						String[] env = new String[3];
						env[0]="TERM="+props.getString("TERM");
						env[1]="PWD="+props.getString("pwd");
						env[2]="HOME="+props.getString("home_directory");

						StringBuffer commandOut = new StringBuffer();
						StringBuffer commandErr = new StringBuffer();
						
						ExecuteCommand exe = new ExecuteCommand(commandString);
						exe.environment = env;
						exe.processOut = commandOut;
						exe.processErr = commandErr;
						exe.waitForProcess = true;
						exe.finishedWaitTime = 1500;
						
						try {
							exe.run();
							
							rval = commandOut.toString()+"\n\n"+commandErr;
						}
						catch(Exception ex) {
							error("Could not execute command: "+commandString,null);
						}
					}
					else {
						String[] env = new String[3];
						env[0]="TERM="+props.getString("TERM");
						env[1]="PWD="+props.getString("pwd");
						env[2]="HOME="+props.getString("home_directory");

						try {
							Runtime r = Runtime.getRuntime();
							Process proc = r.exec(commandString,env);
							Object sync = new Object();
							StreamReader reader1 = new StreamReader(in,proc.getOutputStream(),sync);
							StreamReader reader2 = new StreamReader(proc.getInputStream(),out,sync);
							StreamReader reader3 = new StreamReader(proc.getErrorStream(),out,new Object());
							//							proc.getOutputStream().write(("\n\n\n").getBytes());
							proc.getOutputStream().flush();
							reader1.printStuff = true;
							reader1.start();
							//							reader2.printStuff = true;
							reader2.start();
							//							reader3.printStuff = true;
							reader3.start();
							try {
								synchronized(sync) {
									sync.wait();
								}
							}
							catch(Exception e) {
								e.printStackTrace();
							}
							reader1.stopRunning = true;
							reader2.stopRunning = true;
							reader3.stopRunning = true;
						}
						catch(Exception e) {
							error("Error attempting to execute command: "+commandString,e);
						}
					}
					rval = new String("Attempted to execute command: "+commandString+"\n");
					for(int x=0;x<listeners.size();x++) {
						try {
							listener = (JerminalListenerInterface)listeners.elementAt(x);
							listener.notifyFinishedRunningBinary();
						}
						catch(Exception e) {;}
					}
				}
				else {
					rval = new String("Command "+operator+" not found\n");
					JerminalListenerInterface listener = null;
					for(int x=0;x<listeners.size();x++) {
						try {
							listener = (JerminalListenerInterface)listeners.elementAt(x);
							listener.notifyInvalidCommand();
						}
						catch(Exception e) {;}
					}
				}
			}
		}
		//		catch(UnauthorizedAccessException bleh) {
		//			error("You are not authorized to use that command",null);
		//		}
		catch(Exception e) {
			error("Error in evaluate ",e);
		}


		//		rval = "Operator="+operator;
		//		for(int x=0;x<operands.size();x++) {
		//			rval = rval+" Operand("+x+")="+operands.elementAt(x);
		//		}
		//		rval = rval + "\n";

		return(rval);
	}

	







	/**
	 * This method is responsible for parsing a String of arguments and making evaluations 
	 * based on the syntax. The reserved characters are '\'=escape character, '$'=evaluate 
	 * environment variable, '%'=evaluate the string, '"'=quote groups of arguments, '('=
	 * start of a simple type cast, ')'=ending the cast, and whitespace which is used to delimit
	 * the operand and the operators.
	 *
	 * @param args This is the String of arguments that need to be parsed
	 * @return returns a Vector of argument Objects.
	 *
	 */
	protected Vector parseArgs(String args)
	{
		Vector rval = new Vector();
		
		char c;
		int index = 0;
		boolean dq = false;
		boolean bslash = false;
		boolean dollar = false;
		boolean percent = false;
		boolean openParen = false;
		String word = "";
		String castString = null;

		while(index < args.length()) {
			c = args.charAt(index);
			if(c == '$') {
				if(bslash) {
					word = word+"$";
					bslash = false;
				}
				else {
					dollar = !dollar;
				}
			}
			else if(c == '%') {
				if(bslash) {
					word = word+"%";
					bslash = false;
				}
				else {
					percent = !percent;
				}				
			}
			else if(c == '(') {
				if(bslash) {
					word = word+"(";
					bslash = false;
				}
				else {
					word = word+"(";
					openParen = true;
				}
			}
			else if(c == ')') {
				if(bslash) {
					word = word+")";
					bslash = false;
				}
				else {
					if(openParen) {
						castString = word;
					}
					word = word+")";
					openParen = false;
				}
			}
			else if(c == '\"') {
				if((dq) && (!bslash)) {
					if(dollar && !percent) {
						rval.addElement(props.getProperty(word));
						dollar = false;
					}
					else if(percent && dollar) {
						rval.addElement(evaluate(props.getString(word)));
						percent = false;
						dollar = false;
					}
					else if(percent && !dollar) {
						//System.out.println("EVALUATING "+word);
						rval.addElement(evaluate(word));
						percent = false;
					}
					else {
						rval.addElement(word);
					}

					word = "";
					dq = false;
				}
				else {
					if(bslash) {
						word = word+"\"";
					}
					else if(!dq) {
						if(!word.equals("")) {
							if(dollar && !percent) {
								rval.addElement(props.getProperty(word));
								dollar = false;
							}
							else if(percent && dollar) {
								rval.addElement(evaluate(props.getString(word)));
								percent = false;
								dollar = false;
							}
							else if(percent && !dollar) {
								//System.out.println("EVALUATING "+word);
								rval.addElement(evaluate(word));
								percent = false;
							}
							else {
								rval.addElement(word);
							}
							
							word = "";
						}
						
						dq = true;
					}
				}
				bslash = false;
			}
			// We are assuming here that newlines never get passed in the String
			else if(Character.isWhitespace(c)) {
				if(dq) {
					word = word+c;
				}
				else {
					if(!word.equals("")) {
						if(dollar && !percent) {
							rval.addElement(props.getProperty(word));
							dollar = false;
						}
						else if(percent && dollar) {
							rval.addElement(evaluate(props.getString(word)));
							percent = false;
							dollar = false;
						}
						else if(percent && !dollar) {
							//System.out.println("EVALUATING "+word);
							rval.addElement(evaluate(word));
							percent = false;
						}
						else {
							rval.addElement(word);
						}

						word = "";
					}
				}

				bslash = false;
			}
			else if(c == '\\') {
				//				if(bslash || (!dq)) {
				if(bslash) {
					word = word + "\\";
					bslash = false;
				}
				else {
					bslash = true;
				}
			}
			else {
				if(bslash) {
					if(c == 'n') {
						word = word + "\n";
					}
					else if(c == 't') {
						word = word + "\t";
					}
					else if(c == 'b') {
						word = word + "\b";
					}
					else if(c == 'f') {
						word = word + "\f";
					}
					bslash = false;
				}
				else {
					word = word + c;
					bslash = false;
				}
			}

			index++;
		}

		if(!word.equals("")) {
			if(dollar && !percent) {
				rval.addElement(props.getProperty(word));
				dollar = false;
			}
			else if(percent && dollar) {
				rval.addElement(evaluate(props.getString(word)));
				percent = false;
				dollar = false;
			}
			else if(percent && !dollar) {
				//System.out.println("EVALUATING "+word);
				rval.addElement(evaluate(word));
				percent = false;
			}
			else {
				rval.addElement(word);
			}
		}

		return(rval);
	}



	/**
	 * This prints messages to the System.out as well as the JerminalSession if the
	 * 'verbose' jerminal environment variable variable exists.
	 */
	public void print(String s)
	{
		if(props.hasProperty("verbose")) {
			try {
				localPrintStream.println("JerminalSession: "+s);
				localPrintStream.flush();
			}
			catch(Exception e) {;}
		}
		System.out.println("JerminalSession: "+s);
	}

	/**
	 * This prints error messages to System.err as well as the JerminalSession if the 
	 * 'print_errors' jerminal environment variable exists.
	 */
	public void error(String s, Exception e) 
	{
		if(props.hasProperty("print_errors")) {
			try {
				if(e!=null) {
					localPrintStream.println("JerminalSession Error: "+s+" : "+e);
					if(showStackTracesOnJerm) {
						e.printStackTrace(localPrintStream);
					}
					localPrintStream.flush();
				}
				else {
					localPrintStream.println("JerminalSession Error: "+s);
					localPrintStream.flush();
				}
			}
			catch(Exception sdae) {;}
		}
		if(e!=null) {
			System.err.println("JerminalSession Error: "+s+" : "+e);
			e.printStackTrace(System.err);
		}
		else {
			System.err.println("JerminalSession Error: "+s);
		}
	}
			


	/**
	 * This helper Thread reads from a stream and writes to another. When it is finished, it notifies the syncObject.
	 */
	class StreamReader extends Thread
	{
		public boolean stopRunning;

		public boolean printStuff;

		OutputStream out;
		InputStream in;
		Object sync;

		public StreamReader(InputStream datain, OutputStream dataout, Object syncObject)
		{
			super();
			out = dataout;
			in = datain;
			sync = syncObject;
			stopRunning = false;
		}

		public void run()
		{
			int count = 0;
			try {
				while(!stopRunning) {
					if(in.available()>0) {
						byte[] buffer = new byte[5000];
						int bytesRead = in.read(buffer);

						//						int i = in.read();
						/*
						for(int x=0;x<bytesRead;x++) {
							int i = (int)buffer[x];
							if(i!=-1) { 
								count=0;
								if(printStuff) {
									System.out.println("STREAMREADER: "+i+", "+((char)i));
								}
								out.write(i);
								try {
									out.flush();
								}
								catch(Exception e) {
									e.printStackTrace();
									break;
								}
							}
						}
						*/
						out.write(buffer,0,bytesRead);
						out.flush();
						//						else { 
						//							break; }
					}
					else {
						count++;
						sleep(100);
						if(count>10) {
							break;
						}
					}
				}
			}
			catch(Exception e) {
				if(printStuff) {
					e.printStackTrace();
				}
			}

			try {
				synchronized(sync) {
					sync.notifyAll();
				}
			}
			catch(Exception e) {;}
		}



	}


	public static Vector commaToVector(String in)
	{
		Vector rval = new Vector();

		if(in!=null) {
			try {
				int i = in.indexOf(",");
				while(i!=-1) {
					String tmp = in.substring(0,i);
					rval.addElement(tmp.trim());
					in = in.substring(i+1);
					i = in.indexOf(",");
				}
				if(in.trim().length()>0) {
					rval.addElement(in.trim());
				}
			}
			catch(Exception e) {
				e.printStackTrace();
			}
		}

		return(rval);
	}




	public static void main(String[] args)
	{
		Props p = new Props();

		if(args.length>0) {
			p.loadProps(args[0]);
		}
		JerminalSession j = new JerminalSession(System.in, System.out, p, true,true);
		if(args.length>1) {
			j.evaluate("source "+args[1]);
		}

		j.start();
	}


}
